"""Test the `glue` directives and roles."""

from IPython.core.displaypub import DisplayPublisher
from IPython.core.interactiveshell import InteractiveShell
import nbformat
import pytest

from myst_nb.ext.glue import extract_glue_data, glue


class MockDisplayPublisher(DisplayPublisher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.publish_calls = []

    def publish(self, data, **kwargs):
        kwargs["data"] = data
        self.publish_calls.append(kwargs)


@pytest.fixture()
def mock_ipython():
    """A mock IPython shell for testing notebook cell executions."""
    shell = InteractiveShell.instance()  # type: InteractiveShell
    shell.display_pub = MockDisplayPublisher()
    yield shell.display_pub
    InteractiveShell.clear_instance()


def test_glue_func_text(mock_ipython):
    glue("a", "b")
    assert mock_ipython.publish_calls == [
        {
            "metadata": {"scrapbook": {"name": "a", "mime_prefix": ""}},
            "data": {"text/plain": "'b'"},
        }
    ]


def test_glue_func_obj(mock_ipython):
    class Obj:
        def __repr__(self):
            return "repr"

        def _repr_html_(self):
            return "<p>repr</p>"

    glue("a", Obj())
    assert mock_ipython.publish_calls == [
        {
            "metadata": {"scrapbook": {"name": "a", "mime_prefix": ""}},
            "data": {"text/html": "<p>repr</p>", "text/plain": "repr"},
        }
    ]


def test_glue_func_obj_no_display(mock_ipython):
    class Obj:
        def __repr__(self):
            return "repr"

        def _repr_html_(self):
            return "<p>repr</p>"

    glue("a", Obj(), display=False)
    assert mock_ipython.publish_calls == [
        {
            "metadata": {
                "scrapbook": {
                    "name": "a",
                    "mime_prefix": "application/papermill.record/",
                }
            },
            "data": {
                "application/papermill.record/text/html": "<p>repr</p>",
                "application/papermill.record/text/plain": "repr",
            },
        }
    ]


def test_extract_glue_data(get_test_path):
    path = get_test_path("with_glue.ipynb")
    with open(path) as handle:
        notebook = nbformat.read(handle, as_version=4)
    data = extract_glue_data(notebook, [], None)
    assert set(data) == {
        "key_text1",
        "key_float",
        "key_undisplayed",
        "key_df",
        "key_plt",
        "sym_eq",
    }


@pytest.mark.sphinx_params("with_glue.ipynb", conf={"nb_execution_mode": "off"})
def test_parser(sphinx_run, clean_doctree, file_regression):
    """Test a sphinx build."""
    # TODO test duplicate warning in docutils
    sphinx_run.build()
    # print(sphinx_run.status())
    # print(sphinx_run.warnings())
    assert sphinx_run.warnings() == ""
    assert sphinx_run.env.nb_metadata["with_glue"]["glue"] == [
        "key_text1",
        "key_float",
        "key_undisplayed",
        "key_df",
        "key_plt",
        "sym_eq",
    ]
    doctree = clean_doctree(sphinx_run.get_resolved_doctree("with_glue"))
    file_regression.check(
        doctree.pformat(),
        encoding="utf-8",
    )



# Configuration file for the Sphinx documentation builder.
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import os

# -- Project information -----------------------------------------------------

project = "MyST-NB"
copyright = "2023, Executable Book Project"
author = "Executable Book Project"

master_doc = "index"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_nb",
    "sphinx_copybutton",
    "sphinx_book_theme",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx_design",
    "sphinx_autodoc_typehints",
]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]

myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "html_image",
]

nb_custom_formats = {".Rmd": ["jupytext.reads", {"fmt": "Rmd"}]}
nb_execution_mode = "cache"
nb_execution_show_tb = "READTHEDOCS" in os.environ
nb_execution_timeout = 60  # Note: 30 was timing out on RTD
nb_ipywidgets_js = {
    "https://cdnjs.cloudflare.com/ajax/libs/require.js/2.3.4/require.min.js": {
        "integrity": "sha256-Ae2Vz/4ePdIu6ZyI/5ZGsYnb+m0JlOmKPjt6XZ9JJkA=",
        "crossorigin": "anonymous",
    },
    "https://cdn.jsdelivr.net/npm/@jupyter-widgets/html-manager@*/dist/embed-amd.js": {
        "data-jupyter-widgets-cdn": "https://cdn.jsdelivr.net/npm/",
        "crossorigin": "anonymous",
    },
}
# nb_render_image_options = {"width": "200px"}
# application/vnd.plotly.v1+json and application/vnd.bokehjs_load.v0+json
suppress_warnings = ["mystnb.unknown_mime_type"]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3.8", None),
    "jb": ("https://jupyterbook.org/", None),
    "myst": ("https://myst-parser.readthedocs.io/en/latest/", None),
    "markdown_it": ("https://markdown-it-py.readthedocs.io/en/latest", None),
    "nbclient": ("https://nbclient.readthedocs.io/en/latest", None),
    "nbformat": ("https://nbformat.readthedocs.io/en/latest", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master", None),
}
intersphinx_cache_limit = 5

# ignore these type annotations
nitpick_ignore = [
    ("py:class", klass)
    for klass in [
        "Path",
        "pathlib._local.Path",
        "docutils.nodes.document",
        "docutils.nodes.Node",
        "docutils.nodes.Element",
        "nodes.Element",
        "docutils.nodes.container",
        "docutils.nodes.system_message",
        "DocutilsNbRenderer",
        "SphinxNbRenderer",
        "nbformat.notebooknode.NotebookNode",
        "nbf.NotebookNode",
        "NotebookNode",
        "pygments.lexer.RegexLexer",
        "LoggerType",
        "ExecutionResult",
        "MdParserConfig",
        "NbParserConfig",
        "NbReader",
        # Literal values are not supported
        "typing_extensions.Literal",
        "typing_extensions.Literal[show, remove, remove - warn, warn, error, severe]",
        "off",
        "force",
        "auto",
        "cache",
        "commonmark",
        "gfm",
        "myst",
    ]
]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_title = ""
html_theme = "sphinx_book_theme"
html_logo = "_static/logo-wide.svg"
html_favicon = "_static/logo-square.svg"
html_theme_options = {
    "github_url": "https://github.com/executablebooks/myst-nb",
    "repository_url": "https://github.com/executablebooks/myst-nb",
    "repository_branch": "master",
    "home_page_in_toc": True,
    "path_to_docs": "docs",
    "show_navbar_depth": 1,
    "use_edit_page_button": True,
    "use_repository_button": True,
    "use_download_button": True,
    "launch_buttons": {
        "binderhub_url": "https://mybinder.org",
        "notebook_interface": "classic",
    },
    "navigation_with_keys": False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["custom.css"]

copybutton_selector = "div:not(.output) > div.highlight pre"


def setup(app):
    """Add functions to the Sphinx setup."""
    import subprocess
    from typing import cast

    from docutils import nodes
    from docutils.parsers.rst import directives
    from myst_parser.config.main import MdParserConfig
    from sphinx.application import Sphinx
    from sphinx.util.docutils import SphinxDirective

    from myst_nb.core.config import NbParserConfig, Section

    app = cast(Sphinx, app)

    # this is required to register the coconut kernel with Jupyter,
    # to execute docs/examples/coconut-lang.md
    subprocess.check_call(["coconut", "--jupyter"])

    class _ConfigBase(SphinxDirective):
        """Directive to automate printing of the configuration."""

        @staticmethod
        def table_header():
            return [
                "```````{list-table}",
                ":header-rows: 1",
                "",
                "* - Name",
                "  - Type",
                "  - Default",
                "  - Description",
            ]

        @staticmethod
        def field_default(value):
            default = " ".join(f"{value!r}".splitlines())
            if len(default) > 20:
                default = default[:20] + "..."
            return default

        @staticmethod
        def field_type(field):
            ctype = " ".join(str(field.type).splitlines())
            ctype = ctype.replace("typing.", "")
            ctype = ctype.replace("typing_extensions.", "")
            for tname in ("str", "int", "float", "bool"):
                ctype = ctype.replace(f"<class '{tname}'>", tname)
            return ctype

    class MystNbConfigDirective(_ConfigBase):
        required_arguments = 1
        option_spec = {
            "sphinx": directives.flag,
            "section": lambda x: directives.choice(
                x, ["config", "read", "execute", "render"]
            ),
        }

        def run(self):
            """Run the directive."""
            level_name = directives.choice(
                self.arguments[0], ["global_lvl", "file_lvl", "cell_lvl"]
            )
            level = Section[level_name]

            config = NbParserConfig()
            text = self.table_header()
            count = 0
            for name, value, field in config.as_triple():
                # filter by sphinx options
                if "sphinx" in self.options and field.metadata.get("sphinx_exclude"):
                    continue
                # filter by level
                sections = field.metadata.get("sections") or []
                if level not in sections:
                    continue
                # filter by section
                if "section" in self.options:
                    section = Section[self.options["section"]]
                    if section not in sections:
                        continue

                if level == Section.global_lvl:
                    name = f"nb_{name}"
                elif level == Section.cell_lvl:
                    name = field.metadata.get("cell_key", name)

                description = " ".join(field.metadata.get("help", "").splitlines())
                default = self.field_default(value)
                ctype = self.field_type(field)
                text.extend(
                    [
                        f"* - `{name}`",
                        f"  - `{ctype}`",
                        f"  - `{default}`",
                        f"  - {description}",
                    ]
                )

                count += 1

            if not count:
                return []

            text.append("```````")
            node = nodes.Element()
            self.state.nested_parse(text, 0, node)
            return node.children

    class MystConfigDirective(_ConfigBase):
        option_spec = {
            "sphinx": directives.flag,
        }

        def run(self):
            """Run the directive."""
            config = MdParserConfig()
            text = self.table_header()
            count = 0
            for name, value, field in config.as_triple():
                # filter by sphinx options
                if "sphinx" in self.options and field.metadata.get("sphinx_exclude"):
                    continue

                name = f"myst_{name}"
                description = " ".join(field.metadata.get("help", "").splitlines())
                default = self.field_default(value)
                ctype = self.field_type(field)
                text.extend(
                    [
                        f"* - `{name}`",
                        f"  - `{ctype}`",
                        f"  - `{default}`",
                        f"  - {description}",
                    ]
                )

                count += 1

            if not count:
                return []

            text.append("```````")
            node = nodes.Element()
            self.state.nested_parse(text, 0, node)
            return node.children

    class DocutilsCliHelpDirective(SphinxDirective):
        """Directive to print the docutils CLI help."""

        has_content = False
        required_arguments = 0
        optional_arguments = 0
        final_argument_whitespace = False
        option_spec = {}

        def run(self):
            """Run the directive."""
            import io

            from docutils import nodes
            from docutils.frontend import OptionParser

            from myst_nb.docutils_ import Parser as DocutilsParser

            stream = io.StringIO()
            OptionParser(
                components=(DocutilsParser,),
                usage="mystnb-docutils-<writer> [options] [<source> [<destination>]]",
            ).print_help(stream)
            return [nodes.literal_block("", stream.getvalue())]

    app.add_directive("myst-config", MystConfigDirective)
    app.add_directive("mystnb-config", MystNbConfigDirective)
    app.add_directive("docutils-cli-help", DocutilsCliHelpDirective)



import json
import os
from pathlib import Path
import re
import uuid
import shutil

import bs4
from docutils.nodes import image as image_node
from nbconvert.filters import strip_ansi
from nbdime.diffing.notebooks import (
    diff_notebooks,
    set_notebook_diff_ignores,
    set_notebook_diff_targets,
)
from nbdime.prettyprint import pretty_print_diff
import nbformat as nbf
import pytest
import sphinx
from sphinx import version_info as sphinx_version_info
from sphinx.util.console import nocolor

pytest_plugins = "sphinx.testing.fixtures"

# -Diff Configuration-#
NB_VERSION = 4
set_notebook_diff_ignores({"/nbformat_minor": True})
set_notebook_diff_targets(metadata=False)

TEST_FILE_DIR = Path(__file__).parent.joinpath("notebooks")


@pytest.fixture(autouse=True, scope="session")
def build_matplotlib_font_cache():
    """This is to mitigate errors on CI VMs, where you can get the message:
    "Matplotlib is building the font cache" in output notebooks
    """
    from matplotlib.font_manager import FontManager

    FontManager()


def _split_ext(conf, sphinx_params):
    if custom_formats := conf.get("nb_custom_formats"):
        split_files = [
            file.rstrip(k)
            for file in sphinx_params["files"]
            for k in custom_formats.keys()
            if file.endswith(k)
        ]
    else:
        split_files = [os.path.splitext(file)[0] for file in sphinx_params["files"]]

    return split_files[0], split_files


@pytest.fixture()
def get_test_path():
    def _get_test_path(name):
        return TEST_FILE_DIR.joinpath(name)

    return _get_test_path


class SphinxFixture:
    """A class returned by the ``sphinx_run`` fixture, to run sphinx,
    and retrieve aspects of the build.
    """

    def __init__(self, app, filenames):
        self.app = app
        self.env = app.env
        self.files = filenames
        self.software_versions = (
            f".sphinx{sphinx.version_info[0]}"  # software version tracking for fixtures
        )

        # self.nb_file = nb_file
        # self.nb_name = os.path.splitext(nb_file)[0]

    def build(self):
        """Run the sphinx build."""
        # TODO reset streams before each build,
        # but this was wiping the warnings of a build
        self.app.build()

    def status(self):
        """Return the stdout stream of the sphinx build."""
        return self.app._status.getvalue().strip()

    def warnings(self):
        """Return the stderr stream of the sphinx build."""
        return self.app._warning.getvalue().strip()

    def invalidate_files(self):
        """Invalidate the files, such that it will be flagged for a re-read."""
        for name in self.files:
            self.env.all_docs.pop(name)

    def get_resolved_doctree(self, docname=None):
        """Load and return the built docutils.document, after post-transforms."""
        docname = docname or self.files[0]
        doctree = self.env.get_and_resolve_doctree(docname, self.app.builder)
        doctree["source"] = docname
        return doctree

    def get_doctree(self, docname=None):
        """Load and return the built docutils.document."""
        docname = docname or self.files[0]
        doctree = self.env.get_doctree(docname)
        doctree["source"] = docname
        return doctree

    def get_html(self, index=0):
        """Return the built HTML file."""
        name = self.files[index]
        _path = self.app.outdir / (name + ".html")
        if not _path.exists():
            pytest.fail("html not output")
        return bs4.BeautifulSoup(_path.read_text(), "html.parser")

    def get_nb(self, index=0):
        """Return the output notebook (after any execution)."""
        name = self.files[index]
        _path = self.app.srcdir / "_build" / "jupyter_execute" / (name + ".ipynb")
        if not _path.exists():
            pytest.fail("notebook not output")
        return _path.read_text(encoding="utf-8")

    def get_report_file(self, index=0):
        """Return the report file for a failed execution."""
        name = self.files[index]
        _path = self.app.outdir / "reports" / (name + ".err.log")
        if not _path.exists():
            pytest.fail("report log not output")
        return _path.read_text()


@pytest.fixture()
def sphinx_params(request):
    """Parameters that are specified by 'pytest.mark.sphinx_params'
    are passed to the ``sphinx_run`` fixture::

        @pytest.mark.sphinx_params("name.ipynb", conf={"option": "value"})
        def test_something(sphinx_run):
            ...

    The first file specified here will be set as the master_doc
    """
    markers = request.node.iter_markers("sphinx_params")
    kwargs = {}
    if markers is not None:
        for info in reversed(list(markers)):
            kwargs.update(info.kwargs)
            kwargs["files"] = info.args
    return kwargs


@pytest.fixture()
def sphinx_run(sphinx_params, make_app, tmp_path):
    """A fixture to setup and run a sphinx build, in a sandboxed folder.

    The `myst_nb` extension is added by default,
    and the first file will be set as the masterdoc

    """
    assert len(sphinx_params["files"]) > 0, sphinx_params["files"]
    conf = sphinx_params.get("conf", {})
    buildername = sphinx_params.get("buildername", "html")

    master_doc, split_files = _split_ext(conf, sphinx_params)

    confoverrides = {
        "extensions": ["myst_nb"],
        "master_doc": master_doc,
        "exclude_patterns": ["_build"],
        "nb_execution_show_tb": True,
    }
    confoverrides.update(conf)

    current_dir = os.getcwd()
    if "working_dir" in sphinx_params:
        base_dir = Path(sphinx_params["working_dir"]) / str(uuid.uuid4())
    else:
        base_dir = tmp_path
    srcdir = base_dir / "source"
    srcdir.mkdir(exist_ok=True)
    os.chdir(base_dir)
    (srcdir / "conf.py").write_text(
        "# conf overrides (passed directly to sphinx):\n"
        + "\n".join(
            ["# " + ll for ll in json.dumps(confoverrides, indent=2).splitlines()]
        )
        + "\n"
    )
    if "language" in conf:
        shutil.copytree(TEST_FILE_DIR / "locale", srcdir / "locale")

    for nb_file in sphinx_params["files"]:
        nb_path = TEST_FILE_DIR.joinpath(nb_file)
        assert nb_path.exists(), nb_path
        (srcdir / nb_file).parent.mkdir(exist_ok=True)
        (srcdir / nb_file).write_text(
            nb_path.read_text(encoding="utf-8"), encoding="utf-8"
        )

    nocolor()

    # For compatibility with multiple versions of sphinx, convert pathlib.Path to
    # sphinx.testing.path.path here.
    if sphinx_version_info >= (7, 2):
        app_srcdir = srcdir
    else:
        from sphinx.testing.path import path

        app_srcdir = path(os.fspath(srcdir))
    app = make_app(
        buildername=buildername, srcdir=app_srcdir, confoverrides=confoverrides
    )

    yield SphinxFixture(app, split_files)

    # reset working directory
    os.chdir(current_dir)


def empty_non_deterministic_outputs(cell):
    if "outputs" in cell and len(cell.outputs):
        for item in cell.outputs:
            if "data" in item and "image/png" in item.data:
                item.data["image/png"] = ""
            if "filenames" in item.get("metadata", {}):
                item["metadata"]["filenames"] = {
                    k: os.path.basename(v)
                    for k, v in item["metadata"]["filenames"].items()
                }
            if "traceback" in item:
                item["traceback"] = [strip_ansi(line) for line in item["traceback"]]


@pytest.fixture()
def check_nbs():
    def _check_nbs(obtained_filename, expected_filename):
        obtained_nb = nbf.read(str(obtained_filename), nbf.NO_CONVERT)
        expect_nb = nbf.read(str(expected_filename), nbf.NO_CONVERT)
        obtained_nb.nbformat_minor = 5
        expect_nb.nbformat_minor = 5
        for cell in expect_nb.cells:
            empty_non_deterministic_outputs(cell)
            cell.id = "none"
        for cell in obtained_nb.cells:
            empty_non_deterministic_outputs(cell)
            cell.id = "none"
        diff = diff_notebooks(obtained_nb, expect_nb)
        filename_without_path = str(expected_filename)[
            str(expected_filename).rfind("/") + 1 :
        ]
        if diff:
            raise AssertionError(
                pretty_print_diff(obtained_nb, diff, str(filename_without_path))
            )

    return _check_nbs


@pytest.fixture()
def clean_doctree():
    def _func(doctree):
        if os.name == "nt":  # on Windows file paths are absolute
            findall = getattr(doctree, "findall", doctree.traverse)
            for node in findall(image_node):  # type: image_node
                if "candidates" in node:
                    node["candidates"]["*"] = (
                        "_build/jupyter_execute/"
                        + os.path.basename(node["candidates"]["*"])
                    )
                if "uri" in node:
                    node["uri"] = "_build/jupyter_execute/" + os.path.basename(
                        node["uri"]
                    )
        return doctree

    return _func


# comparison files will need updating
# alternatively the resolution of https://github.com/ESSS/pytest-regressions/issues/32
@pytest.fixture()
def file_regression(file_regression):
    return FileRegression(file_regression)


class FileRegression:
    ignores = (
        # TODO: Remove when support for Sphinx<=6 is dropped,
        re.escape(" translation_progress=\"{'total': 0, 'translated': 0}\""),
        # TODO: Remove when support for Sphinx<7.2 is dropped,
        r"original_uri=\"[^\"]*\"\s",
        # TODO: Remove when support for Sphinx<8 is dropped,
        re.escape(' translated="True"'),
        re.escape(" translation_progress=\"{'total': 4, 'translated': 2}\""),
    )

    def __init__(self, file_regression):
        self.file_regression = file_regression

    def check(self, data, **kwargs):
        return self.file_regression.check(self._strip_ignores(data), **kwargs)

    def _strip_ignores(self, data):
        for ig in self.ignores:
            data = re.sub(ig, "", data)
        return data



"""The sphinx parser implementation for myst-nb."""

from __future__ import annotations

from collections import defaultdict
from html import escape
import json
from pathlib import Path
import re
from typing import Any, DefaultDict, cast

from docutils import nodes
from markdown_it.token import Token
from markdown_it.tree import SyntaxTreeNode
from myst_parser.config.main import MdParserConfig, merge_file_level
from myst_parser.mdit_to_docutils.base import token_line
from myst_parser.mdit_to_docutils.sphinx_ import SphinxRenderer
from myst_parser.parsers.mdit import create_md_parser
from myst_parser.parsers.sphinx_ import MystParser
import nbformat
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment
from sphinx.environment.collectors import EnvironmentCollector
from sphinx.transforms.post_transforms import SphinxPostTransform
from sphinx.util import logging as sphinx_logging
from sphinx.util.docutils import SphinxTranslator

from myst_nb._compat import findall
from myst_nb.core.config import NbParserConfig
from myst_nb.core.execute import ExecutionResult, create_client
from myst_nb.core.loggers import DEFAULT_LOG_TYPE, SphinxDocLogger
from myst_nb.core.nb_to_tokens import nb_node_to_dict, notebook_to_tokens
from myst_nb.core.read import create_nb_reader
from myst_nb.core.render import (
    MditRenderMixin,
    MimeData,
    NbElementRenderer,
    create_figure_context,
    get_mime_priority,
    load_renderer,
)
from myst_nb.warnings_ import MystNBWarnings, create_warning

SPHINX_LOGGER = sphinx_logging.getLogger(__name__)


class SphinxEnvType(BuildEnvironment):
    """Sphinx build environment, including attributes set by myst_nb."""

    myst_config: MdParserConfig
    mystnb_config: NbParserConfig
    nb_metadata: DefaultDict[str, dict]
    nb_new_exec_data: bool


class Parser(MystParser):
    """Sphinx parser for Jupyter Notebook formats, containing MyST Markdown."""

    supported = ("myst-nb",)
    translate_section_name = None

    config_section = "myst-nb parser"
    config_section_dependencies = ("parsers",)

    env: SphinxEnvType

    def parse(self, inputstring: str, document: nodes.document) -> None:
        """Parse source text.

        :param inputstring: The source string to parse
        :param document: The root docutils node to add AST elements to
        """
        assert self.env is not None, "env not set"
        document_path = self.env.doc2path(self.env.docname)

        # get a logger for this document
        logger = SphinxDocLogger(document)

        # get markdown parsing configuration
        md_config: MdParserConfig = self.env.myst_config

        # get notebook rendering configuration
        nb_config: NbParserConfig = self.env.mystnb_config

        # create a reader for the notebook
        nb_reader = create_nb_reader(document_path, md_config, nb_config, inputstring)
        # If the nb_reader is None, then we default to a standard Markdown parser
        if nb_reader is None or document.current_source.endswith("<translated>"):
            return super().parse(inputstring, document)
        notebook = nb_reader.read(inputstring)

        # update the global markdown config with the file-level config
        warning = lambda wtype, msg: create_warning(  # noqa: E731
            document, msg, line=1, append_to=document, subtype=wtype
        )
        nb_reader.md_config = merge_file_level(
            nb_reader.md_config, notebook.metadata, warning
        )

        # potentially replace kernel name with alias
        kernel_name = notebook.metadata.get("kernelspec", {}).get("name", None)
        if kernel_name is not None and nb_config.kernel_rgx_aliases:
            for rgx, alias in nb_config.kernel_rgx_aliases.items():
                if re.fullmatch(rgx, kernel_name):
                    logger.debug(
                        f"Replaced kernel name: {kernel_name!r} -> {alias!r}",
                        subtype="kernel",
                    )
                    notebook.metadata["kernelspec"]["name"] = alias
                    break

        # Update mystnb configuration with notebook level metadata
        if nb_config.metadata_key in notebook.metadata:
            overrides = nb_node_to_dict(notebook.metadata[nb_config.metadata_key])
            overrides.pop("output_folder", None)  # this should not be overridden
            try:
                nb_config = nb_config.copy(**overrides)
            except Exception as exc:
                logger.warning(
                    f"Failed to update configuration with notebook metadata: {exc}",
                    subtype="config",
                )
            else:
                logger.debug(
                    "Updated configuration with notebook metadata", subtype="config"
                )

        # Setup the parser
        mdit_parser = create_md_parser(nb_reader.md_config, SphinxNbRenderer)
        mdit_parser.options["document"] = document
        mdit_parser.options["nb_config"] = nb_config
        mdit_renderer: SphinxNbRenderer = mdit_parser.renderer  # type: ignore
        mdit_env: dict[str, Any] = {}

        # load notebook element renderer class from entry-point name
        # this is separate from SphinxNbRenderer, so that users can override it
        renderer_name = nb_config.render_plugin
        nb_renderer: NbElementRenderer = load_renderer(renderer_name)(
            mdit_renderer, logger
        )
        # we temporarily store nb_renderer on the document,
        # so that roles/directives can access it
        document.attributes["nb_renderer"] = nb_renderer
        # we currently do this early, so that the nb_renderer has access to things
        mdit_renderer.setup_render(mdit_parser.options, mdit_env)  # type: ignore

        # parse notebook structure to markdown-it tokens
        # note, this does not assume that the notebook has been executed yet
        mdit_tokens = notebook_to_tokens(notebook, mdit_parser, mdit_env, logger)

        # open the notebook execution client,
        # this may execute the notebook immediately or during the page render
        with create_client(
            notebook, document_path, nb_config, logger, nb_reader.read_fmt
        ) as nb_client:
            mdit_parser.options["nb_client"] = nb_client
            # convert to docutils AST, which is added to the document
            mdit_renderer.render(mdit_tokens, mdit_parser.options, mdit_env)

        # save final execution data
        if nb_client.exec_metadata:
            NbMetadataCollector.set_exec_data(
                self.env, self.env.docname, nb_client.exec_metadata
            )
            if nb_client.exec_metadata["traceback"]:
                # store error traceback in outdir and log its path
                reports_file = Path(self.env.app.outdir).joinpath(
                    "reports", *(self.env.docname + ".err.log").split("/")
                )
                reports_file.parent.mkdir(parents=True, exist_ok=True)
                reports_file.write_text(
                    nb_client.exec_metadata["traceback"], encoding="utf8"
                )
                logger.warning(
                    f"Notebook exception traceback saved in: {reports_file}",
                    subtype="exec",
                )

        # write final (updated) notebook to output folder (utf8 is standard encoding)
        path = self.env.docname.split("/")
        ipynb_path = path[:-1] + [path[-1] + ".ipynb"]
        content = nbformat.writes(notebook).encode("utf-8")
        nb_renderer.write_file(ipynb_path, content, overwrite=True)

        # write glue data to the output folder,
        # and store the keys to environment doc metadata,
        # so that they may be used in any post-transform steps
        if nb_client.glue_data:
            glue_path = path[:-1] + [path[-1] + ".glue.json"]
            nb_renderer.write_file(
                glue_path,
                json.dumps(nb_client.glue_data, cls=BytesEncoder).encode("utf8"),
                overwrite=True,
            )
            NbMetadataCollector.set_doc_data(
                self.env, self.env.docname, "glue", list(nb_client.glue_data.keys())
            )

        # move some document metadata to environment metadata,
        # so that we can later read it from the environment,
        # rather than having to load the whole doctree
        for key, (uri, kwargs) in document.attributes.pop("nb_js_files", {}).items():
            NbMetadataCollector.add_js_file(
                self.env, self.env.docname, key, uri, kwargs
            )

        # remove temporary state
        document.attributes.pop("nb_renderer")


class SphinxNbRenderer(SphinxRenderer, MditRenderMixin):
    """A sphinx renderer for Jupyter Notebooks."""

    def render_nb_initialise(self, token: SyntaxTreeNode) -> None:
        env = cast(BuildEnvironment, self.sphinx_env)
        metadata = self.nb_client.nb_metadata
        special_keys = ["kernelspec", "language_info", "source_map"]
        for key in special_keys:
            if key in metadata:
                # save these special keys on the metadata, rather than as docinfo
                # note, sphinx_book_theme checks kernelspec is in the metadata
                env.metadata[env.docname][key] = metadata.get(key)

        # forward the remaining metadata to the front_matter renderer
        special_keys.append("widgets")
        top_matter = {k: v for k, v in metadata.items() if k not in special_keys}
        self.render_front_matter(
            Token(  # type: ignore
                "front_matter",
                "",
                0,
                map=[0, 0],
                content=top_matter,  # type: ignore[arg-type]
            ),
        )

    def _render_nb_cell_code_outputs(
        self, token: SyntaxTreeNode, outputs: list[nbformat.NotebookNode]
    ) -> None:
        """Render a notebook code cell's outputs."""
        line = token_line(token, 0)
        cell_index = token.meta["index"]
        metadata = token.meta["metadata"]
        # render the outputs
        for output_index, output in enumerate(outputs):
            if output.output_type == "stream":
                if output.name == "stdout":
                    _nodes = self.nb_renderer.render_stdout(
                        output, metadata, cell_index, line
                    )
                    self.add_line_and_source_path_r(_nodes, token)
                    self.current_node.extend(_nodes)
                elif output.name == "stderr":
                    _nodes = self.nb_renderer.render_stderr(
                        output, metadata, cell_index, line
                    )
                    self.add_line_and_source_path_r(_nodes, token)
                    self.current_node.extend(_nodes)
                else:
                    pass  # TODO warning
            elif output.output_type == "error":
                _nodes = self.nb_renderer.render_error(
                    output, metadata, cell_index, line
                )
                self.add_line_and_source_path_r(_nodes, token)
                self.current_node.extend(_nodes)
            elif output.output_type in ("display_data", "execute_result"):
                # Note, this is different to the docutils implementation,
                # where we directly select a single output, based on the mime_priority.
                # Here, we do not know the mime priority until we know the output format
                # so we output all the outputs during this parsing phase
                # (this is what sphinx caches as "output format agnostic" AST),
                # and replace the mime_bundle with the format specific output
                # in a post-transform (run per output format on the cached AST)

                figure_options = (
                    self.get_cell_level_config(
                        "render_figure_options", metadata, line=line
                    )
                    or None
                )

                with create_figure_context(self, figure_options, line):
                    mime_bundle = nodes.container(nb_element="mime_bundle")
                    with self.current_node_context(mime_bundle):
                        for mime_type, data in output["data"].items():
                            mime_container = nodes.container(mime_type=mime_type)
                            with self.current_node_context(mime_container):
                                _nodes = self.nb_renderer.render_mime_type(
                                    MimeData(
                                        mime_type,
                                        data,
                                        cell_metadata=metadata,
                                        output_metadata=output.get("metadata", {}),
                                        cell_index=cell_index,
                                        output_index=output_index,
                                        line=line,
                                    )
                                )
                                self.current_node.extend(_nodes)
                            if mime_container.children:
                                self.current_node.append(mime_container)
                    if mime_bundle.children:
                        self.add_line_and_source_path_r([mime_bundle], token)
                        self.current_node.append(mime_bundle)
            else:
                create_warning(
                    self.document,
                    f"Unsupported output type: {output.output_type}",
                    line=line,
                    append_to=self.current_node,
                    # wtype=DEFAULT_LOG_TYPE,
                    subtype=MystNBWarnings.OUTPUT_TYPE,
                )


class SelectMimeType(SphinxPostTransform):
    """Select the mime type to render from mime bundles,
    based on the builder and its associated priority list.
    """

    default_priority = 4  # TODO set correct priority

    def run(self, **kwargs: Any) -> None:
        """Run the transform."""
        # get priority list for this builder
        # TODO allow for per-notebook/cell priority dicts?
        bname = self.app.builder.name
        priority_list = get_mime_priority(
            bname, self.config["nb_mime_priority_overrides"]
        )

        def condition(node):
            return (
                isinstance(node, nodes.container)
                and node.attributes.get("nb_element", "") == "mime_bundle"
            )

        # remove/replace_self will not work with an iterator
        for node in list(findall(self.document)(condition)):
            # get available mime types
            mime_types = [node["mime_type"] for node in node.children]
            if not mime_types:
                node.parent.remove(node)
                continue
            # select top priority
            index = None
            for mime_type in priority_list:
                try:
                    index = mime_types.index(mime_type)
                except ValueError:
                    continue
                else:
                    break
            if index is None:
                mime_string = ",".join(repr(m) for m in mime_types)
                SPHINX_LOGGER.warning(
                    f"No mime type available in priority list for builder {bname!r} "
                    f"({mime_string}) [{DEFAULT_LOG_TYPE}.mime_priority]",
                    type=DEFAULT_LOG_TYPE,
                    subtype="mime_priority",
                    location=node,
                )
                node.parent.remove(node)
            elif not node.children[index].children:
                node.parent.remove(node)
            else:
                node.replace_self(node.children[index].children)


class NbMetadataCollector(EnvironmentCollector):
    """Collect myst-nb specific metadata, and handle merging of parallel builds."""

    @staticmethod
    def set_doc_data(env: SphinxEnvType, docname: str, key: str, value: Any) -> None:
        """Add nb metadata for a docname to the environment."""
        if not hasattr(env, "nb_metadata"):
            env.nb_metadata = defaultdict(dict)
        env.nb_metadata.setdefault(docname, {})[key] = value

    @staticmethod
    def get_doc_data(env: SphinxEnvType) -> DefaultDict[str, dict]:
        """Get myst-nb docname -> metadata dict."""
        if not hasattr(env, "nb_metadata"):
            env.nb_metadata = defaultdict(dict)
        return env.nb_metadata

    @classmethod
    def set_exec_data(
        cls, env: SphinxEnvType, docname: str, value: ExecutionResult
    ) -> None:
        """Add nb metadata for a docname to the environment."""
        cls.set_doc_data(env, docname, "exec_data", value)
        # TODO this does not take account of cache data
        cls.note_exec_update(env)

    @classmethod
    def get_exec_data(cls, env: SphinxEnvType, docname: str) -> ExecutionResult | None:
        """Get myst-nb docname -> execution data."""
        return cls.get_doc_data(env)[docname].get("exec_data")

    def get_outdated_docs(  # type: ignore[override]
        self,
        app: Sphinx,
        env: SphinxEnvType,
        added: set[str],
        changed: set[str],
        removed: set[str],
    ) -> list[str]:
        # called before any docs are read
        env.nb_new_exec_data = False
        return []

    @staticmethod
    def note_exec_update(env: SphinxEnvType) -> None:
        """Note that a notebook has been executed."""
        env.nb_new_exec_data = True

    @staticmethod
    def new_exec_data(env: SphinxEnvType) -> bool:
        """Return whether any notebooks have updated execution data."""
        return getattr(env, "nb_new_exec_data", False)

    @classmethod
    def add_js_file(
        cls,
        env: SphinxEnvType,
        docname: str,
        key: str,
        uri: str | None,
        kwargs: dict[str, str],
    ):
        """Register a JavaScript file to include in the HTML output."""
        if not hasattr(env, "nb_metadata"):
            env.nb_metadata = defaultdict(dict)
        js_files = env.nb_metadata.setdefault(docname, {}).setdefault("js_files", {})
        # TODO handle whether overrides are allowed
        js_files[key] = (uri, kwargs)

    @classmethod
    def get_js_files(
        cls, env: SphinxEnvType, docname: str
    ) -> dict[str, tuple[str | None, dict[str, str]]]:
        """Get myst-nb docname -> execution data."""
        return cls.get_doc_data(env)[docname].get("js_files", {})

    def clear_doc(  # type: ignore[override]
        self,
        app: Sphinx,
        env: SphinxEnvType,
        docname: str,
    ) -> None:
        if not hasattr(env, "nb_metadata"):
            env.nb_metadata = defaultdict(dict)
        env.nb_metadata.pop(docname, None)

    def process_doc(self, app: Sphinx, doctree: nodes.document) -> None:
        pass

    def merge_other(  # type: ignore[override]
        self,
        app: Sphinx,
        env: SphinxEnvType,
        docnames: set[str],
        other: SphinxEnvType,
    ) -> None:
        if not hasattr(env, "nb_metadata"):
            env.nb_metadata = defaultdict(dict)
        other_metadata = getattr(other, "nb_metadata", defaultdict(dict))
        for docname in docnames:
            env.nb_metadata[docname] = other_metadata[docname]
        if other.nb_new_exec_data:
            env.nb_new_exec_data = True


class BytesEncoder(json.JSONEncoder):
    """A JSON encoder that accepts b64 (and other *ascii*) bytestrings."""

    def default(self, obj):
        if isinstance(obj, bytes):
            return obj.decode("ascii")
        return json.JSONEncoder.default(self, obj)


class HideCodeCellNode(nodes.Element):
    """Node for hiding cell input."""

    @classmethod
    def add_to_app(cls, app: Sphinx):
        app.add_node(cls, html=(visit_HideCellInput, depart_HideCellInput))


def visit_HideCellInput(self: SphinxTranslator, node: HideCodeCellNode):
    classes = " ".join(node["classes"])
    self.body.append(f'<details class="admonition hide {classes}">\n')
    self.body.append('<summary aria-label="Toggle hidden content">\n')
    self.body.append(
        f'<p class="collapsed admonition-title">{escape(node["prompt_show"])}</p>\n'
    )
    self.body.append(
        f'<p class="expanded admonition-title">{escape(node["prompt_hide"])}</p>\n'
    )
    self.body.append("</summary>\n")


def depart_HideCellInput(self: SphinxTranslator, node: HideCodeCellNode):
    self.body.append("</details>\n")


class HideInputCells(SphinxPostTransform):
    """Hide input cells in the HTML output."""

    default_priority = 199
    formats = ("html",)

    def run(self, **kwargs):
        for node in findall(self.document)(nodes.container):
            if (
                node.get("nb_element") == "cell_code"
                and node.get("hide_mode")
                and node.children
            ):
                hide_mode = node.get("hide_mode")
                has_input = node.children[0].get("nb_element") == "cell_code_source"
                has_output = node.children[-1].get("nb_element") == "cell_code_output"

                if has_input and hide_mode == "all":
                    # wrap everything and place a summary above the input
                    wrap_node = HideCodeCellNode(
                        prompt_show=node["prompt_show"].replace("{type}", "content"),
                        prompt_hide=node["prompt_hide"].replace("{type}", "content"),
                    )
                    wrap_node["classes"].append("above-input")
                    wrap_node.extend(node.children)
                    node.children = [wrap_node]

                if has_input and has_output and hide_mode in ("output", "input+output"):
                    node.children[0]["classes"].append("above-output-prompt")

                if has_input and hide_mode in ("input", "input+output"):
                    # wrap just the input and place a summary above the input
                    wrap_node = HideCodeCellNode(
                        prompt_show=node["prompt_show"].replace("{type}", "source"),
                        prompt_hide=node["prompt_hide"].replace("{type}", "source"),
                    )
                    wrap_node["classes"].append("above-input")
                    code = node.children[0]
                    wrap_node.append(code)
                    node.replace(code, wrap_node)

                if has_input and has_output and hide_mode in ("output", "input+output"):
                    # wrap just the output and place a summary below the input
                    wrap_node = HideCodeCellNode(
                        prompt_show=node["prompt_show"].replace("{type}", "output"),
                        prompt_hide=node["prompt_hide"].replace("{type}", "output"),
                    )
                    wrap_node["classes"].append("below-input")
                    output = node.children[-1]
                    wrap_node.append(output)
                    node.replace(output, wrap_node)

                if (
                    (not has_input)
                    and has_output
                    and hide_mode in ("all", "input+output", "output")
                ):
                    # wrap just the output and place a summary above the output
                    wrap_node = HideCodeCellNode(
                        prompt_show=node["prompt_show"].replace("{type}", "outputs"),
                        prompt_hide=node["prompt_hide"].replace("{type}", "outputs"),
                    )
                    wrap_node["classes"].append("above-output")
                    output = node.children[-1]
                    wrap_node.append(output)
                    node.replace(output, wrap_node)



"""Tests for rendering code cell outputs."""

import pytest

from myst_nb.core.render import EntryPointError, load_renderer
from sphinx.util.fileutil import copy_asset_file


def test_load_renderer_not_found():
    """Test that an error is raised when the renderer is not found."""
    with pytest.raises(EntryPointError, match="No Entry Point found"):
        load_renderer("other")


# TODO sometimes fails in full tests
# def test_load_renderer_not_subclass(monkeypatch):
#     """Test that an error is raised when the renderer is not a subclass."""
#     from importlib_metadata import EntryPoint
#     monkeypatch.setattr(EntryPoint, "load", lambda self: object)
#     with pytest.raises(EntryPointError, match="Entry Point .* not a subclass"):
#         load_renderer("default")


@pytest.mark.sphinx_params("basic_run.ipynb", conf={"nb_execution_mode": "off"})
def test_basic_run(sphinx_run, file_regression):
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    doctree = sphinx_run.get_resolved_doctree("basic_run")
    file_regression.check(doctree.pformat(), extension=".xml", encoding="utf-8")


@pytest.mark.sphinx_params("file_level_config.md")
def test_file_level_config_md(sphinx_run, file_regression):
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    doctree = sphinx_run.get_resolved_doctree("file_level_config")
    file_regression.check(doctree.pformat(), extension=".xml", encoding="utf-8")


@pytest.mark.sphinx_params("file_level_config.ipynb")
def test_file_level_config_ipynb(sphinx_run, file_regression):
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    doctree = sphinx_run.get_resolved_doctree("file_level_config")
    file_regression.check(doctree.pformat(), extension=".xml", encoding="utf-8")


@pytest.mark.sphinx_params("complex_outputs.ipynb", conf={"nb_execution_mode": "off"})
def test_complex_outputs(sphinx_run, clean_doctree, file_regression):
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    doctree = clean_doctree(sphinx_run.get_resolved_doctree("complex_outputs"))
    file_regression.check(
        doctree.pformat().replace(".jpeg", ".jpg"), extension=".xml", encoding="utf-8"
    )


@pytest.mark.sphinx_params(
    "complex_outputs.ipynb",
    conf={"nb_execution_mode": "off"},
    buildername="latex",
)
def test_complex_outputs_latex(sphinx_run, clean_doctree, file_regression):
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    doctree = clean_doctree(sphinx_run.get_resolved_doctree("complex_outputs"))
    file_regression.check(
        doctree.pformat().replace(".jpeg", ".jpg"), extension=".xml", encoding="utf-8"
    )


@pytest.mark.sphinx_params(
    "basic_stderr.ipynb",
    conf={"nb_execution_mode": "off", "nb_output_stderr": "remove"},
)
def test_stderr_remove(sphinx_run, file_regression):
    """Test configuring all stderr outputs to be removed."""
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    doctree = sphinx_run.get_resolved_doctree("basic_stderr")
    file_regression.check(doctree.pformat(), extension=".xml", encoding="utf-8")


@pytest.mark.sphinx_params("basic_stderr.ipynb", conf={"nb_execution_mode": "off"})
def test_stderr_tag(sphinx_run, file_regression):
    """Test configuring stderr outputs to be removed from a single cell,
    using `remove-stderr` in the `cell.metadata.tags`.
    """
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    doctree = sphinx_run.get_resolved_doctree("basic_stderr")
    file_regression.check(doctree.pformat(), extension=".xml", encoding="utf-8")


@pytest.mark.sphinx_params(
    "merge_streams.ipynb",
    conf={"nb_execution_mode": "off", "nb_merge_streams": True},
)
def test_merge_streams(sphinx_run, file_regression):
    """Test configuring multiple concurrent stdout/stderr outputs to be merged."""
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    doctree = sphinx_run.get_resolved_doctree("merge_streams")
    file_regression.check(doctree.pformat(), extension=".xml", encoding="utf-8")


@pytest.mark.sphinx_params(
    "merge_streams_parallel.ipynb",
    conf={"nb_execution_mode": "off", "nb_merge_streams": True},
)
def test_merge_streams_parallel(sphinx_run, file_regression):
    """Test configuring multiple concurrent stdout/stderr outputs to be merged."""
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    doctree = sphinx_run.get_resolved_doctree("merge_streams_parallel")
    file_regression.check(doctree.pformat(), extension=".xml", encoding="utf-8")


@pytest.mark.sphinx_params(
    "metadata_image.ipynb",
    conf={"nb_execution_mode": "off", "nb_cell_metadata_key": "myst"},
)
def test_metadata_image(sphinx_run, clean_doctree, file_regression):
    """Test configuring image attributes to be rendered from cell metadata."""
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    doctree = clean_doctree(sphinx_run.get_resolved_doctree("metadata_image"))
    file_regression.check(doctree.pformat(), extension=".xml", encoding="utf8")


@pytest.mark.sphinx_params(
    "metadata_multiple_image.ipynb",
    conf={"nb_execution_mode": "off", "nb_cell_metadata_key": "myst"},
)
def test_metadata_multiple_image(sphinx_run, clean_doctree, file_regression):
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    doctree = clean_doctree(sphinx_run.get_resolved_doctree("metadata_multiple_image"))
    file_regression.check(doctree.pformat(), extension=".xml", encoding="utf8")


@pytest.mark.sphinx_params(
    "metadata_image_output.ipynb",
    conf={"nb_execution_mode": "force"},
)
def test_metadata_image_output(
    sphinx_run, clean_doctree, file_regression, get_test_path
):
    """Test configuring image attributes to be rendered from cell metadata."""
    asset_path = get_test_path("example.jpg")
    copy_asset_file(str(asset_path), str(sphinx_run.app.srcdir))
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    doctree = clean_doctree(sphinx_run.get_resolved_doctree("metadata_image_output"))
    file_regression.check(
        doctree.pformat().replace(".jpeg", ".jpg"), extension=".xml", encoding="utf-8"
    )


@pytest.mark.sphinx_params(
    "metadata_figure.ipynb",
    conf={"nb_execution_mode": "off", "nb_cell_metadata_key": "myst"},
)
def test_metadata_figure(sphinx_run, clean_doctree, file_regression):
    """Test configuring figure attributes to be rendered from cell metadata."""
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    doctree = clean_doctree(sphinx_run.get_resolved_doctree("metadata_figure"))
    doctree_string = doctree.pformat()
    # change, presumably with new docutils version
    doctree_string = doctree_string.replace(
        '<figure ids="fun-fish" names="fun-fish">',
        '<figure align="default" ids="fun-fish" names="fun-fish">',
    )
    file_regression.check(
        doctree_string.replace(".jpeg", ".jpg"), extension=".xml", encoding="utf-8"
    )


@pytest.mark.sphinx_params("unknown_mimetype.ipynb", conf={"nb_execution_mode": "off"})
def test_unknown_mimetype(sphinx_run, file_regression):
    """Test that unknown mimetypes provide a warning."""
    sphinx_run.build()
    warning = "skipping unknown output mime type: unknown [mystnb.unknown_mime_type]"
    assert warning in sphinx_run.warnings()
    doctree = sphinx_run.get_resolved_doctree("unknown_mimetype")
    file_regression.check(doctree.pformat(), extension=".xml", encoding="utf-8")


@pytest.mark.sphinx_params("hide_cell_content.ipynb", conf={"nb_execution_mode": "off"})
def test_hide_cell_content(sphinx_run, file_regression):
    """Test that hiding cell contents produces the correct AST."""
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    doctree = sphinx_run.get_resolved_doctree("hide_cell_content")
    file_regression.check(doctree.pformat(), extension=".xml", encoding="utf-8")


@pytest.mark.sphinx_params(
    "scroll_outputs.ipynb", conf={"nb_execution_mode": "off", "nb_scroll_outputs": True}
)
def test_scroll_outputs(sphinx_run, file_regression):
    """Test that scrollable outputs are rendered correctly."""
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    doctree = sphinx_run.get_resolved_doctree("scroll_outputs")
    file_regression.check(doctree.pformat(), extension=".xml", encoding="utf-8")



"""Run parsing tests against the docutils parser."""

from io import StringIO
import json
from pathlib import Path

from docutils.core import publish_doctree, publish_string
import pytest
import sphinx
import yaml

from myst_nb.docutils_ import Parser

FIXTURE_PATH = Path(__file__).parent.joinpath("nb_fixtures")


@pytest.mark.param_file(FIXTURE_PATH / "basic.txt")
def test_basic(file_params):
    """Test basic parsing."""
    if (
        "Footnote definitions defined in different cells" in file_params.title
        and sphinx.version_info[0] < 5
    ):
        pytest.skip("footnote definition ids changes")
    dct = yaml.safe_load(file_params.content)
    dct.update({"nbformat": 4, "nbformat_minor": 4})
    dct.setdefault("metadata", {})
    dct["metadata"].setdefault(
        "kernelspec", {"name": "python3", "display_name": "Python 3", "language": ""}
    )
    report_stream = StringIO()
    doctree = publish_doctree(
        json.dumps(dct),
        parser=Parser(),
        settings_overrides={
            "nb_execution_mode": "off",
            "nb_output_folder": "",
            "myst_all_links_external": True,
            "warning_stream": report_stream,
        },
    )
    assert report_stream.getvalue().rstrip() == ""

    file_params.assert_expected(doctree.pformat(), rstrip=True)


@pytest.mark.param_file(FIXTURE_PATH / "reporter_warnings.txt")
def test_reporting(file_params):
    """Test that warnings and errors are reported as expected."""
    dct = yaml.safe_load(file_params.content)
    dct.update({"metadata": {}, "nbformat": 4, "nbformat_minor": 4})
    report_stream = StringIO()
    publish_doctree(
        json.dumps(dct),
        parser=Parser(),
        settings_overrides={
            "nb_execution_mode": "off",
            "nb_output_folder": "",
            "warning_stream": report_stream,
        },
    )
    file_params.assert_expected(report_stream.getvalue(), rstrip=True)


def test_html_resources(tmp_path):
    """Test HTML resources are correctly output."""
    report_stream = StringIO()
    result = publish_string(
        json.dumps({"cells": [], "metadata": {}, "nbformat": 4, "nbformat_minor": 4}),
        parser=Parser(),
        writer_name="html",
        settings_overrides={
            "nb_execution_mode": "off",
            "nb_output_folder": str(tmp_path),
            "warning_stream": report_stream,
            "output_encoding": "unicode",
            "embed_stylesheet": False,
        },
    )
    assert report_stream.getvalue().rstrip() == ""
    assert "mystnb.css" in result
    assert "pygments.css" in result
    assert tmp_path.joinpath("mystnb.css").is_file()
    assert tmp_path.joinpath("pygments.css").is_file()



"""Test the `eval` directives and roles."""

import pytest


@pytest.mark.sphinx_params("with_eval.md", conf={"nb_execution_mode": "inline"})
def test_sphinx(sphinx_run, clean_doctree, file_regression):
    """Test a sphinx build."""
    sphinx_run.build()
    # print(sphinx_run.status())
    # print(sphinx_run.warnings())
    assert sphinx_run.warnings() == ""
    doctree = clean_doctree(sphinx_run.get_resolved_doctree("with_eval"))
    file_regression.check(
        doctree.pformat(),
        encoding="utf-8",
    )



"""Test parsing of already executed notebooks."""

import os
from pathlib import Path

import pytest


@pytest.mark.sphinx_params("basic_run.ipynb", conf={"nb_execution_mode": "off"})
def test_basic_run(sphinx_run, file_regression):
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    assert set(sphinx_run.env.metadata["basic_run"].keys()) == {
        "test_name",
        "wordcount",
        "kernelspec",
        "language_info",
    }
    assert set(sphinx_run.env.nb_metadata["basic_run"].keys()) == set()
    assert sphinx_run.env.metadata["basic_run"]["test_name"] == "notebook1"
    assert sphinx_run.env.metadata["basic_run"]["kernelspec"] == {
        "display_name": "Python 3",
        "language": "python",
        "name": "python3",
    }
    file_regression.check(
        sphinx_run.get_doctree().pformat(), extension=".xml", encoding="utf-8"
    )

    filenames = {
        p.name
        for p in Path(
            os.fspath(sphinx_run.app.srcdir / "_build" / "jupyter_execute")
        ).iterdir()
    }
    assert filenames == {"basic_run.ipynb"}


@pytest.mark.sphinx_params(
    "basic_run_intl.ipynb", conf={"language": "es", "locale_dirs": ["locale"]}
)
def test_basic_run_intl(sphinx_run, file_regression):
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    assert set(sphinx_run.env.metadata["basic_run_intl"].keys()) == {
        "test_name",
        "wordcount",
        "kernelspec",
        "language_info",
    }
    assert set(sphinx_run.env.nb_metadata["basic_run_intl"].keys()) == set()
    assert sphinx_run.env.metadata["basic_run_intl"]["test_name"] == "notebook1"
    assert sphinx_run.env.metadata["basic_run_intl"]["kernelspec"] == {
        "display_name": "Python 3",
        "language": "python",
        "name": "python3",
    }
    file_regression.check(
        sphinx_run.get_doctree().pformat(), extension=".xml", encoding="utf-8"
    )

    filenames = {
        p.name
        for p in Path(
            os.fspath(sphinx_run.app.srcdir / "_build" / "jupyter_execute")
        ).iterdir()
    }
    assert filenames == {"basic_run_intl.ipynb"}


@pytest.mark.sphinx_params("complex_outputs.ipynb", conf={"nb_execution_mode": "off"})
def test_complex_outputs(sphinx_run, file_regression):
    sphinx_run.build()
    assert sphinx_run.warnings() == ""

    assert set(sphinx_run.env.metadata["complex_outputs"].keys()) == {
        "ipub",
        "hide_input",
        "nav_menu",
        "celltoolbar",
        "latex_envs",
        "jupytext",
        "toc",
        "varInspector",
        "wordcount",
        "kernelspec",
        "language_info",
    }
    assert set(sphinx_run.env.nb_metadata["complex_outputs"].keys()) == set()
    assert sphinx_run.env.metadata["complex_outputs"]["celltoolbar"] == "Edit Metadata"
    assert sphinx_run.env.metadata["complex_outputs"]["hide_input"] == "False"
    assert sphinx_run.env.metadata["complex_outputs"]["kernelspec"] == {
        "display_name": "Python 3",
        "language": "python",
        "name": "python3",
    }
    doctree_string = sphinx_run.get_doctree().pformat()
    if os.name == "nt":  # on Windows image file paths are absolute
        doctree_string = doctree_string.replace(
            Path(sphinx_run.app.srcdir).as_posix() + "/", ""
        )
    file_regression.check(doctree_string, extension=".xml", encoding="utf-8")

    filenames = {
        p.name.replace(".jpeg", ".jpg")
        for p in Path(
            os.fspath(sphinx_run.app.srcdir / "_build" / "jupyter_execute")
        ).iterdir()
    }
    # print(filenames)
    assert filenames == {
        "16832f45917c1c9862c50f0948f64a498402d6ccde1f3a291da17f240797b160.png",
        "a4c9580c74dacf6f3316a3bd2e2a347933aa4463834dcf1bb8f20b4fcb476ae1.jpg",
        "8c43e5c8cccf697754876b7fec1b0a9b731d7900bb585e775a5fa326b4de8c5a.png",
        "complex_outputs.ipynb",
    }


@pytest.mark.sphinx_params(
    "latex_build/index.ipynb",
    "latex_build/other.ipynb",
    conf={"nb_execution_mode": "off"},
    buildername="latex",
)
def test_toctree_in_ipynb(sphinx_run, file_regression):
    sphinx_run.build()
    print(sphinx_run.status())
    print(sphinx_run.warnings())
    file_regression.check(
        sphinx_run.get_doctree("latex_build/other").pformat(), extension=".xml"
    )
    assert sphinx_run.warnings() == ""


@pytest.mark.sphinx_params("ipywidgets.ipynb", conf={"nb_execution_mode": "off"})
def test_ipywidgets(sphinx_run):
    """Test that ipywidget state is extracted and JS is included in the HTML head."""
    sphinx_run.build()
    # print(sphinx_run.status())
    assert sphinx_run.warnings() == ""
    assert "js_files" in sphinx_run.env.nb_metadata["ipywidgets"]
    assert set(sphinx_run.env.nb_metadata["ipywidgets"]["js_files"]) == {
        "ipywidgets_state",
        "ipywidgets_0",
        "ipywidgets_1",
    }
    head_scripts = sphinx_run.get_html().select("head > script")
    assert any("require.js" in script.get("src", "") for script in head_scripts)
    assert any("embed-amd.js" in script.get("src", "") for script in head_scripts)



"""Test the quickstart CLI"""

import os
from pathlib import Path

import nbformat
from sphinx import version_info as sphinx_version_info

from myst_nb.cli import md_to_nb, quickstart


def test_quickstart(tmp_path: Path, make_app):
    """Test the quickstart CLI builds a valid sphinx project."""
    project_path = tmp_path / "project"
    quickstart([str(project_path)])
    assert {p.name for p in project_path.iterdir()} == {
        ".gitignore",
        "conf.py",
        "index.md",
        "notebook1.ipynb",
        "notebook2.md",
    }

    # For compatibility with multiple versions of sphinx, convert pathlib.Path to
    # sphinx.testing.path.path here.
    if sphinx_version_info >= (7, 2):
        app_srcdir = project_path
    else:
        from sphinx.testing.path import path

        app_srcdir = path(os.fspath(project_path))

    app = make_app(srcdir=app_srcdir, buildername="html")
    app.build()
    assert app._warning.getvalue().strip() == ""
    assert (project_path / "_build/html/index.html").exists()


def test_md_to_nb(tmp_path: Path):
    """Test the md_to_nb CLI."""
    path = tmp_path / "notebook.md"
    outpath = path.with_suffix(".ipynb")
    path.write_text(
        """\
---
kernelspec:
    name: python3
---
# Title
+++
next cell
""",
        "utf-8",
    )
    md_to_nb([str(path)])
    assert path.exists()
    with outpath.open("r") as handle:
        nb = nbformat.read(handle, as_version=4)
    assert nb.metadata == {"kernelspec": {"display_name": "python3", "name": "python3"}}
    assert len(nb.cells) == 2



"""Central handling of warnings for the myst-nb extension."""

from __future__ import annotations

from enum import Enum
from typing import Sequence

from docutils import nodes
from myst_parser.warnings_ import MystWarnings
from myst_parser.warnings_ import create_warning as myst_parser_create_warnings

__all__ = [
    "MystWarnings",
    "MystNBWarnings",
    "create_warning",
]


class MystNBWarnings(Enum):
    """MySTNB warning types."""

    LEXER = "lexer"
    """Issue resolving lexer"""

    FIG_CAPTION = "fig_caption"
    """Issue resoliving figure caption"""

    MIME_TYPE = "mime_type"
    """Issue resolving MIME type"""
    OUTPUT_TYPE = "output_type"
    """Issue resolving Output type"""

    CELL_METADATA_KEY = "cell_metadata_key"
    """Issue with a key in a cell's `metadata` dictionary."""
    CELL_CONFIG = "cell_config"
    """Issue with a cell's configuration or metadata."""


def _is_suppressed_warning(
    type: str, subtype: str, suppress_warnings: Sequence[str]
) -> bool:
    """Check whether the warning is suppressed or not.

    Mirrors:
    https://github.com/sphinx-doc/sphinx/blob/47d9035bca9e83d6db30a0726a02dc9265bd66b1/sphinx/util/logging.py
    """
    if type is None:
        return False

    subtarget: str | None

    for warning_type in suppress_warnings:
        if "." in warning_type:
            target, subtarget = warning_type.split(".", 1)
        else:
            target, subtarget = warning_type, None

        if target == type and subtarget in (None, subtype, "*"):
            return True

    return False


def create_warning(
    document: nodes.document,
    message: str,
    subtype: MystNBWarnings | MystWarnings,
    *,
    line: int | None = None,
    append_to: nodes.Element | None = None,
) -> nodes.system_message | None:
    """Generate a warning, logging if it is necessary.

    If the warning type is listed in the ``suppress_warnings`` configuration,
    then ``None`` will be returned and no warning logged.
    """
    # Pass off Myst Parser warnings to that package
    if isinstance(subtype, MystWarnings):
        myst_parser_create_warnings(
            document=document,
            message=message,
            subtype=subtype,
            line=line,
            append_to=append_to,
        )

    wtype = "myst-nb"
    # figure out whether to suppress the warning, if sphinx is available,
    # it will have been set up by the Sphinx environment,
    # otherwise we will use the configuration set by docutils
    suppress_warnings: Sequence[str] = []
    try:
        suppress_warnings = document.settings.env.app.config.suppress_warnings
    except AttributeError:
        suppress_warnings = document.settings.myst_suppress_warnings or []
    if _is_suppressed_warning(wtype, subtype.value, suppress_warnings):
        return None

    kwargs = {"line": line} if line is not None else {}
    message = f"{message} [{wtype}.{subtype.value}]"
    msg_node = document.reporter.warning(message, **kwargs)
    if append_to is not None:
        append_to.append(msg_node)
    return msg_node



"""Test sphinx builds which execute notebooks."""

import os
from pathlib import Path

from IPython import version_info as ipy_version
import pytest

from myst_nb.core.execute import ExecutionError
from myst_nb.sphinx_ import NbMetadataCollector


def regress_nb_doc(file_regression, sphinx_run, check_nbs):
    try:
        file_regression.check(
            sphinx_run.get_nb(),
            check_fn=check_nbs,
            extension=".ipynb",
            encoding="utf-8",
        )
    finally:
        doctree_string = sphinx_run.get_doctree().pformat()
        # TODO this is a difference in the hashing on the CI,
        # with complex_outputs_unrun.ipynb equation PNG, after execution
        # sympy
        doctree_string = doctree_string.replace(
            "91b3db0f47514d451e7c0f5a501d31c4e101b7112050634ad0788405f417782a",
            "e2dfbe330154316cfb6f3186e8f57fc4df8aee03b0303ed1345fc22cd51f66de",
        )
        doctree_string = doctree_string.replace(
            "438c56ea3dcf99d86cd64df1b23e2b436afb25846434efb1cfec7b660ef01127",
            "e2dfbe330154316cfb6f3186e8f57fc4df8aee03b0303ed1345fc22cd51f66de",
        )
        if os.name == "nt":  # on Windows image file paths are absolute
            doctree_string = doctree_string.replace(
                Path(sphinx_run.app.srcdir).as_posix() + "/", ""
            )
        file_regression.check(doctree_string, extension=".xml", encoding="utf-8")


@pytest.mark.sphinx_params("basic_unrun.ipynb", conf={"nb_execution_mode": "auto"})
def test_basic_unrun_auto(sphinx_run, file_regression, check_nbs):
    sphinx_run.build()
    # print(sphinx_run.status())
    assert sphinx_run.warnings() == ""
    assert "test_name" in sphinx_run.app.env.metadata["basic_unrun"]
    regress_nb_doc(file_regression, sphinx_run, check_nbs)

    assert NbMetadataCollector.new_exec_data(sphinx_run.env)
    data = NbMetadataCollector.get_exec_data(sphinx_run.env, "basic_unrun")
    assert data
    assert data["method"] == "auto"
    assert data["succeeded"] is True


@pytest.mark.sphinx_params("basic_unrun.ipynb", conf={"nb_execution_mode": "cache"})
def test_basic_unrun_cache(sphinx_run, file_regression, check_nbs):
    """The outputs should be populated."""
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    assert "test_name" in sphinx_run.app.env.metadata["basic_unrun"]
    regress_nb_doc(file_regression, sphinx_run, check_nbs)

    assert NbMetadataCollector.new_exec_data(sphinx_run.env)
    data = NbMetadataCollector.get_exec_data(sphinx_run.env, "basic_unrun")
    assert data
    assert data["method"] == "cache"
    assert data["succeeded"] is True


@pytest.mark.sphinx_params("basic_unrun.ipynb", conf={"nb_execution_mode": "inline"})
def test_basic_unrun_inline(sphinx_run, file_regression, check_nbs):
    """The outputs should be populated."""
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    assert "test_name" in sphinx_run.app.env.metadata["basic_unrun"]
    regress_nb_doc(file_regression, sphinx_run, check_nbs)

    assert NbMetadataCollector.new_exec_data(sphinx_run.env)
    data = NbMetadataCollector.get_exec_data(sphinx_run.env, "basic_unrun")
    assert data
    assert data["method"] == "inline"
    assert data["succeeded"] is True


@pytest.mark.sphinx_params("basic_unrun.ipynb", conf={"nb_execution_mode": "cache"})
def test_rebuild_cache(sphinx_run):
    """The notebook should only be executed once."""
    sphinx_run.build()
    assert NbMetadataCollector.new_exec_data(sphinx_run.env)
    sphinx_run.invalidate_files()
    sphinx_run.build()
    assert "Using cached" in sphinx_run.status()


@pytest.mark.sphinx_params("basic_unrun.ipynb", conf={"nb_execution_mode": "force"})
def test_rebuild_force(sphinx_run):
    """The notebook should be executed twice."""
    sphinx_run.build()
    assert NbMetadataCollector.new_exec_data(sphinx_run.env)
    sphinx_run.invalidate_files()
    sphinx_run.build()
    assert NbMetadataCollector.new_exec_data(sphinx_run.env)


@pytest.mark.sphinx_params(
    "basic_unrun.ipynb",
    conf={
        "nb_execution_mode": "cache",
        "nb_execution_excludepatterns": ["basic_*"],
    },
)
def test_exclude_path(sphinx_run, file_regression):
    """The notebook should not be executed."""
    sphinx_run.build()
    assert not NbMetadataCollector.new_exec_data(sphinx_run.env)
    assert "Executing" not in sphinx_run.status(), sphinx_run.status()
    file_regression.check(
        sphinx_run.get_doctree().pformat(), extension=".xml", encoding="utf-8"
    )


@pytest.mark.skipif(ipy_version[0] < 8, reason="Error message changes for ipython v8")
@pytest.mark.sphinx_params("basic_failing.ipynb", conf={"nb_execution_mode": "cache"})
def test_basic_failing_cache(sphinx_run, file_regression, check_nbs):
    sphinx_run.build()
    # print(sphinx_run.warnings())
    assert "Executing notebook failed" in sphinx_run.warnings()
    regress_nb_doc(file_regression, sphinx_run, check_nbs)

    data = NbMetadataCollector.get_exec_data(sphinx_run.env, "basic_failing")
    assert data
    assert data["method"] == "cache"
    assert data["succeeded"] is False
    sphinx_run.get_report_file()


@pytest.mark.skipif(ipy_version[0] < 8, reason="Error message changes for ipython v8")
@pytest.mark.sphinx_params("basic_failing.ipynb", conf={"nb_execution_mode": "auto"})
def test_basic_failing_auto(sphinx_run, file_regression, check_nbs):
    sphinx_run.build()
    assert "Executing notebook failed" in sphinx_run.warnings()
    regress_nb_doc(file_regression, sphinx_run, check_nbs)

    data = NbMetadataCollector.get_exec_data(sphinx_run.env, "basic_failing")
    assert data
    assert data["method"] == "auto"
    assert data["succeeded"] is False
    assert data["traceback"]
    sphinx_run.get_report_file()


@pytest.mark.skipif(ipy_version[0] < 8, reason="Error message changes for ipython v8")
@pytest.mark.sphinx_params("basic_failing.ipynb", conf={"nb_execution_mode": "inline"})
def test_basic_failing_inline(sphinx_run, file_regression, check_nbs):
    sphinx_run.build()
    assert "Executing notebook failed" in sphinx_run.warnings()
    regress_nb_doc(file_regression, sphinx_run, check_nbs)

    data = NbMetadataCollector.get_exec_data(sphinx_run.env, "basic_failing")
    assert data
    assert data["method"] == "inline"
    assert data["succeeded"] is False
    assert data["traceback"]
    sphinx_run.get_report_file()


@pytest.mark.skipif(ipy_version[0] < 8, reason="Error message changes for ipython v8")
@pytest.mark.sphinx_params(
    "basic_failing.ipynb",
    conf={"nb_execution_mode": "cache", "nb_execution_allow_errors": True},
)
def test_allow_errors_cache(sphinx_run, file_regression, check_nbs):
    sphinx_run.build()
    # print(sphinx_run.status())
    assert not sphinx_run.warnings()
    regress_nb_doc(file_regression, sphinx_run, check_nbs)


@pytest.mark.skipif(ipy_version[0] < 8, reason="Error message changes for ipython v8")
@pytest.mark.sphinx_params(
    "basic_failing.ipynb",
    conf={"nb_execution_mode": "auto", "nb_execution_allow_errors": True},
)
def test_allow_errors_auto(sphinx_run, file_regression, check_nbs):
    sphinx_run.build()
    # print(sphinx_run.status())
    assert not sphinx_run.warnings()
    regress_nb_doc(file_regression, sphinx_run, check_nbs)


@pytest.mark.sphinx_params(
    "basic_failing.ipynb",
    conf={"nb_execution_raise_on_error": True, "nb_execution_mode": "force"},
)
def test_raise_on_error_force(sphinx_run):
    with pytest.raises(ExecutionError, match="basic_failing.ipynb"):
        sphinx_run.build()


@pytest.mark.sphinx_params(
    "basic_failing.ipynb",
    conf={"nb_execution_raise_on_error": True, "nb_execution_mode": "cache"},
)
def test_raise_on_error_cache(sphinx_run):
    with pytest.raises(ExecutionError, match="basic_failing.ipynb"):
        sphinx_run.build()


@pytest.mark.sphinx_params(
    "complex_outputs_unrun.ipynb", conf={"nb_execution_mode": "cache"}
)
def test_complex_outputs_unrun_cache(sphinx_run, file_regression, check_nbs):
    sphinx_run.build()
    # print(sphinx_run.status())
    assert sphinx_run.warnings() == ""
    regress_nb_doc(file_regression, sphinx_run, check_nbs)

    # Widget view and widget state should make it into the HTML
    scripts = sphinx_run.get_html().select("script")
    assert any(
        "application/vnd.jupyter.widget-view+json" in script.get("type", "")
        for script in scripts
    )
    assert any(
        "application/vnd.jupyter.widget-state+json" in script.get("type", "")
        for script in scripts
    )


@pytest.mark.sphinx_params(
    "complex_outputs_unrun.ipynb", conf={"nb_execution_mode": "auto"}
)
def test_complex_outputs_unrun_auto(sphinx_run, file_regression, check_nbs):
    sphinx_run.build()
    # print(sphinx_run.status())
    assert sphinx_run.warnings() == ""
    regress_nb_doc(file_regression, sphinx_run, check_nbs)

    # Widget view and widget state should make it into the HTML
    scripts = sphinx_run.get_html().select("script")
    assert any(
        "application/vnd.jupyter.widget-view+json" in script.get("type", "")
        for script in scripts
    )
    assert any(
        "application/vnd.jupyter.widget-state+json" in script.get("type", "")
        for script in scripts
    )


@pytest.mark.sphinx_params("basic_unrun.ipynb", conf={"nb_execution_mode": "off"})
def test_no_execute(sphinx_run, file_regression, check_nbs):
    sphinx_run.build()
    # print(sphinx_run.status())
    assert sphinx_run.warnings() == ""
    regress_nb_doc(file_regression, sphinx_run, check_nbs)


@pytest.mark.sphinx_params("basic_unrun.ipynb", conf={"nb_execution_mode": "cache"})
def test_jupyter_cache_path(sphinx_run, file_regression, check_nbs):
    sphinx_run.build()
    assert "Cached executed notebook" in sphinx_run.status()
    assert sphinx_run.warnings() == ""
    regress_nb_doc(file_regression, sphinx_run, check_nbs)


# Testing relative paths within the notebook
@pytest.mark.sphinx_params("basic_relative.ipynb", conf={"nb_execution_mode": "cache"})
def test_relative_path_cache(sphinx_run):
    sphinx_run.build()
    assert "Execution Failed" not in sphinx_run.status(), sphinx_run.status()


@pytest.mark.sphinx_params("basic_relative.ipynb", conf={"nb_execution_mode": "force"})
def test_relative_path_force(sphinx_run):
    sphinx_run.build()
    assert "Execution Failed" not in sphinx_run.status(), sphinx_run.status()


@pytest.mark.sphinx_params(
    "kernel_alias.md",
    conf={"nb_execution_mode": "force", "nb_kernel_rgx_aliases": {"oth.+": "python3"}},
)
def test_kernel_rgx_aliases(sphinx_run):
    sphinx_run.build()
    assert sphinx_run.warnings() == ""


@pytest.mark.sphinx_params(
    "sleep_10.ipynb",
    conf={"nb_execution_mode": "cache", "nb_execution_timeout": 1},
)
def test_execution_timeout(sphinx_run):
    """execution should fail given the low timeout value"""
    sphinx_run.build()
    # print(sphinx_run.warnings())
    assert "Executing notebook failed" in sphinx_run.warnings()


@pytest.mark.sphinx_params(
    "sleep_10_metadata_timeout.ipynb",
    conf={"nb_execution_mode": "cache", "nb_execution_timeout": 60},
)
def test_execution_metadata_timeout(sphinx_run):
    """notebook timeout metadata has higher preference then execution_timeout config"""
    sphinx_run.build()
    # print(sphinx_run.warnings())
    assert "Executing notebook failed" in sphinx_run.warnings()


@pytest.mark.sphinx_params(
    "nb_exec_table.md",
    conf={"nb_execution_mode": "auto"},
)
def test_nb_exec_table(sphinx_run, file_regression):
    """Test that the table gets output into the HTML,
    including a row for the executed notebook.
    """
    sphinx_run.build()
    # print(sphinx_run.status())
    assert not sphinx_run.warnings()
    file_regression.check(
        sphinx_run.get_doctree().pformat(), extension=".xml", encoding="utf-8"
    )
    # print(sphinx_run.get_html())
    rows = sphinx_run.get_html().select("table.docutils tr")
    assert any("nb_exec_table" in row.text for row in rows)


@pytest.mark.sphinx_params(
    "custom-formats.Rmd",
    conf={
        "nb_execution_mode": "auto",
        "nb_custom_formats": {".Rmd": ["jupytext.reads", {"fmt": "Rmd"}]},
    },
)
def test_custom_convert_auto(sphinx_run, file_regression, check_nbs):
    sphinx_run.build()
    # print(sphinx_run.status())
    assert sphinx_run.warnings() == ""
    regress_nb_doc(file_regression, sphinx_run, check_nbs)

    assert NbMetadataCollector.new_exec_data(sphinx_run.env)
    data = NbMetadataCollector.get_exec_data(sphinx_run.env, "custom-formats")
    assert data
    assert data["method"] == "auto"
    assert data["succeeded"] is True


@pytest.mark.sphinx_params(
    "custom-formats.Rmd",
    conf={
        "nb_execution_mode": "cache",
        "nb_custom_formats": {".Rmd": ["jupytext.reads", {"fmt": "Rmd"}]},
    },
)
def test_custom_convert_cache(sphinx_run, file_regression, check_nbs):
    """The outputs should be populated."""
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    regress_nb_doc(file_regression, sphinx_run, check_nbs)

    assert NbMetadataCollector.new_exec_data(sphinx_run.env)
    data = NbMetadataCollector.get_exec_data(sphinx_run.env, "custom-formats")
    assert data
    assert data["method"] == "cache"
    assert data["succeeded"] is True


@pytest.mark.sphinx_params(
    "custom-formats2.extra.exnt",
    conf={
        "nb_execution_mode": "auto",
        "nb_custom_formats": {".extra.exnt": ["jupytext.reads", {"fmt": "Rmd"}]},
    },
)
def test_custom_convert_multiple_extensions_auto(
    sphinx_run, file_regression, check_nbs
):
    """The outputs should be populated."""
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    regress_nb_doc(file_regression, sphinx_run, check_nbs)

    assert NbMetadataCollector.new_exec_data(sphinx_run.env)
    data = NbMetadataCollector.get_exec_data(sphinx_run.env, "custom-formats2")
    assert data
    assert data["method"] == "auto"
    assert data["succeeded"] is True


@pytest.mark.sphinx_params(
    "custom-formats2.extra.exnt",
    conf={
        "nb_execution_mode": "cache",
        "nb_custom_formats": {".extra.exnt": ["jupytext.reads", {"fmt": "Rmd"}]},
    },
)
def test_custom_convert_multiple_extensions_cache(
    sphinx_run, file_regression, check_nbs
):
    """The outputs should be populated."""
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    regress_nb_doc(file_regression, sphinx_run, check_nbs)

    assert NbMetadataCollector.new_exec_data(sphinx_run.env)
    data = NbMetadataCollector.get_exec_data(sphinx_run.env, "custom-formats2")
    assert data
    assert data["method"] == "cache"
    assert data["succeeded"] is True



"""Test notebooks containing code cells with the `load` option."""

import pytest
from sphinx.util.fileutil import copy_asset_file


@pytest.mark.sphinx_params(
    "mystnb_codecell_file.md",
    conf={"nb_execution_mode": "cache", "source_suffix": {".md": "myst-nb"}},
)
def test_codecell_file(sphinx_run, file_regression, check_nbs, get_test_path):
    asset_path = get_test_path("mystnb_codecell_file.py")
    copy_asset_file(str(asset_path), str(sphinx_run.app.srcdir))
    sphinx_run.build()
    assert sphinx_run.warnings() == ""
    assert set(sphinx_run.env.metadata["mystnb_codecell_file"].keys()) == {
        "jupytext",
        "author",
        "source_map",
        "wordcount",
        "kernelspec",
        "language_info",
    }
    assert set(sphinx_run.env.nb_metadata["mystnb_codecell_file"].keys()) == {
        "exec_data",
    }
    assert sphinx_run.env.metadata["mystnb_codecell_file"]["author"] == "Matt"
    assert sphinx_run.env.metadata["mystnb_codecell_file"]["kernelspec"] == {
        "display_name": "Python 3",
        "language": "python",
        "name": "python3",
    }
    try:
        file_regression.check(
            sphinx_run.get_nb(),
            check_fn=check_nbs,
            extension=".ipynb",
            encoding="utf-8",
        )
    finally:
        file_regression.check(
            sphinx_run.get_doctree().pformat(), extension=".xml", encoding="utf-8"
        )


@pytest.mark.sphinx_params(
    "mystnb_codecell_file_warnings.md",
    conf={"nb_execution_mode": "force", "source_suffix": {".md": "myst-nb"}},
)
def test_codecell_file_warnings(sphinx_run, file_regression, check_nbs, get_test_path):
    asset_path = get_test_path("mystnb_codecell_file.py")
    copy_asset_file(str(asset_path), str(sphinx_run.app.srcdir))
    sphinx_run.build()
    # assert (
    #     "mystnb_codecell_file_warnings.md:14 content of code-cell "
    #     "is being overwritten by :load: mystnb_codecell_file.py"
    #     in sphinx_run.warnings()
    # )
    assert set(sphinx_run.env.metadata["mystnb_codecell_file_warnings"].keys()) == {
        "jupytext",
        "author",
        "source_map",
        "wordcount",
        "kernelspec",
        "language_info",
    }
    assert set(sphinx_run.env.nb_metadata["mystnb_codecell_file_warnings"].keys()) == {
        "exec_data",
    }
    assert (
        sphinx_run.env.metadata["mystnb_codecell_file_warnings"]["author"] == "Aakash"
    )
    assert sphinx_run.env.metadata["mystnb_codecell_file_warnings"]["kernelspec"] == {
        "display_name": "Python 3",
        "language": "python",
        "name": "python3",
    }
    try:
        file_regression.check(
            sphinx_run.get_nb(),
            check_fn=check_nbs,
            extension=".ipynb",
            encoding="utf-8",
        )
    finally:
        file_regression.check(
            sphinx_run.get_doctree().pformat(), extension=".xml", encoding="utf-8"
        )



import pytest


@pytest.mark.sphinx_params(
    "basic_unrun.md",
    conf={"nb_execution_mode": "cache", "source_suffix": {".md": "myst-nb"}},
)
def test_basic_run(sphinx_run, file_regression, check_nbs):
    sphinx_run.build()
    # print(sphinx_run.status())
    assert sphinx_run.warnings() == ""
    assert set(sphinx_run.env.metadata["basic_unrun"].keys()) == {
        "file_format",
        "author",
        "source_map",
        "wordcount",
        "kernelspec",
        "language_info",
    }
    assert set(sphinx_run.env.nb_metadata["basic_unrun"].keys()) == {
        "exec_data",
    }
    assert sphinx_run.env.metadata["basic_unrun"]["author"] == "Chris"
    assert sphinx_run.env.metadata["basic_unrun"]["kernelspec"] == {
        "display_name": "Python 3",
        "language": "python",
        "name": "python3",
    }
    file_regression.check(
        sphinx_run.get_nb(), check_fn=check_nbs, extension=".ipynb", encoding="utf-8"
    )
    file_regression.check(
        sphinx_run.get_doctree().pformat(), extension=".xml", encoding="utf-8"
    )


@pytest.mark.sphinx_params(
    "basic_unrun.md",
    conf={"nb_execution_mode": "off", "source_suffix": {".md": "myst-nb"}},
)
def test_basic_run_exec_off(sphinx_run, file_regression, check_nbs):
    sphinx_run.build()
    # print(sphinx_run.status())
    assert set(sphinx_run.env.metadata["basic_unrun"].keys()) == {
        "file_format",
        "author",
        "source_map",
        "wordcount",
        "kernelspec",
    }
    assert set(sphinx_run.env.nb_metadata["basic_unrun"].keys()) == set()
    assert sphinx_run.env.metadata["basic_unrun"]["author"] == "Chris"

    file_regression.check(
        sphinx_run.get_nb(), check_fn=check_nbs, extension=".ipynb", encoding="utf-8"
    )
    file_regression.check(
        sphinx_run.get_doctree().pformat(), extension=".xml", encoding="utf-8"
    )


@pytest.mark.sphinx_params(
    "basic_nometadata.md",
    conf={"nb_execution_mode": "off", "source_suffix": {".md": "myst-nb"}},
)
def test_basic_nometadata(sphinx_run):
    """A myst-markdown notebook with no jupytext metadata should raise a warning."""
    sphinx_run.build()
    # print(sphinx_run.status())
    assert "Found an unexpected `code-cell`" in sphinx_run.warnings()



from pygments.token import Text, Token
import pytest

from myst_nb.core import lexers


@pytest.mark.parametrize(
    ("bold", "faint", "fg_color", "bg_color", "expected"),
    (
        (False, False, False, False, Text),
        (True, False, False, False, Token.Color.Bold),
        (True, False, "Red", False, Token.Color.Bold.Red),
        (True, False, "Red", "Blue", Token.Color.Bold.Red.BGBlue),
        (True, True, "Red", "Blue", Token.Color.Bold.Faint.Red.BGBlue),
    ),
)
def test_token_from_lexer_state(bold, faint, fg_color, bg_color, expected):
    ret = lexers._token_from_lexer_state(bold, faint, fg_color, bg_color)
    assert ret == expected


def _highlight(text):
    return tuple(lexers.AnsiColorLexer().get_tokens(text))


def test_plain_text():
    assert _highlight("hello world\n") == ((Text, "hello world\n"),)


def test_simple_colors():
    assert _highlight(
        "plain text\n"
        "\x1b[31mred text\n"
        "\x1b[1;32mbold green text\n"
        "\x1b[39mfg color turned off\n"
        "\x1b[0mplain text after reset\n"
        "\x1b[1mbold text\n"
        "\x1b[43mbold from previous line with yellow bg\n"
        "\x1b[49mbg color turned off\n"
        "\x1b[2mfaint turned on\n"
        "\x1b[22mbold turned off\n"
    ) == (
        (Text, "plain text\n"),
        (Token.Color.Red, "red text\n"),
        (Token.Color.Bold.Green, "bold green text\n"),
        (Token.Color.Bold, "fg color turned off\n"),
        (Text, "plain text after reset\n"),
        (Token.Color.Bold, "bold text\n"),
        (Token.Color.Bold.BGYellow, "bold from previous line with yellow bg\n"),
        (Token.Color.Bold, "bg color turned off\n"),
        (Token.Color.Bold.Faint, "faint turned on\n"),
        (Text, "bold turned off\n"),
    )


def test_highlight_empty_end_specifier():
    ret = _highlight("plain\x1b[31mred\x1b[mplain\n")
    assert ret == ((Text, "plain"), (Token.Color.Red, "red"), (Text, "plain\n"))


def test_ignores_unrecognized_ansi_color_codes():
    """It should just strip and ignore any unrecognized color ANSI codes."""
    assert _highlight(
        # unknown int code
        "\x1b[99m"
        "plain text\n"
        # invalid non-int code
        "\x1b[=m"
        "plain text\n"
    ) == (
        (Text, "plain text\n"),
        (Text, "plain text\n"),
    )


def test_ignores_valid_ansi_non_color_codes():
    """It should just strip and ignore any non-color ANSI codes.

    These include things like moving the cursor position, erasing lines, etc.
    """
    assert _highlight(
        # restore cursor position
        "\x1b[u"
        "plain "
        # move cursor backwards 55 steps
        "\x1b[55C"
        "text\n"
    ) == (
        # Ideally these would be just one token, but our regex isn't smart
        # enough yet.
        (Text, "plain "),
        (Text, "text\n"),
    )


def test_ignores_completely_invalid_escapes():
    """It should strip and ignore invalid escapes.

    This shouldn't happen in valid ANSI text, but we could have an escape
    followed by garbage.
    """
    assert _highlight("plain \x1b[%text\n") == (
        (Text, "plain "),
        (Text, "%text\n"),
    )
